#ifndef	lint
char	*Rcsid = "$Header: /u/moraes/src/unshar/RCS/unshar.c,v 1.2 1992/05/06 03:44:37 root Exp moraes $";
#endif

/*
 * getmaps
 *
 * Get the net maps from USENET as published by Karen and Mark Horton, in
 * "shar" format. Because of paranoia the sh is not used but instead a DFA
 * recognizing the appropriate commands.
 *
 * Lee Ward		10/13/84
 */
/* Modified a lot at University of Toronto by Mark Moraes, Dec 1990 */
/*
 * TODO:  make the backup procedure optional -- differentiate between
 * verbose logs and warnings!!  log() should be on stdout, warning()
 * on stderr.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef DIRENT
# include <dirent.h>
# ifndef direct
#  define direct dirent
# endif
#else
# include <sys/dir.h>
#endif

#include <errno.h>

/*
 * when DEBUG is defined, we malloc buffers instead of using the stack,
 * so we can use a debugging malloc to check for overflow
 */
#ifdef DEBUG
# define MAXSTR 128	/* easier to test overflow cases */
#else
# define MAXSTR 1024
#endif

#define SAVEHDRS	"UNSHAR.HDR"

/* STREQN is an optimised strncmp(a,b,n)==0; assumes n > 0 */
#define STREQN(a, b, n) ((a)[0] == (b)[0] && strncmp(a, b, n) == 0)

char *progname;
static char *mapgrp;
static char *mapdir;
static char *seqfil = "./.seq";
static char *savfil = SAVEHDRS;
static char *badchars = "* \\'\"`[]{}<>;?&~#\t\n";
static FILE *savefp;
static int nosubdirs;
static int debug;

/* forwards */
extern void domaps(), error(), log(), logerr(), archdir();
extern char *getwrd(), *getsed();
extern int mkmaps(), gotcmd();

/* For getopt() */
extern int optind;
extern char *optarg;

extern int errno;
extern int atoi();
extern char *malloc();
extern time_t time();
extern char *ctime();
extern char *mktemp();
extern int scandir();

int
main(argc, argv)
int argc;
char *argv[];
{
	int	c, errs = 0, x, savehdrs = 0;

	progname = argv[0] ? argv[0] : "(no-argv[0])";
	while((c = getopt(argc, argv, "g:c:snfd")) != EOF) {
		switch (c) {
		case 'g':
			mapgrp = optarg;
			break;
		case 'c':
			mapdir = optarg;
			break;
		case 's':
			savehdrs++;
			break;
		case 'n':
			savehdrs = 0;
			break;
		case 'f':
			nosubdirs++;
			break;
		case 'd':
			debug++;
			break;
		default:
			errs++;
			break;
		}
	}
	if (errs > 0 || (mapgrp != NULL && optind < argc)) {
		(void) fprintf(stderr,
			       "Usage: %s [-g newsgroup] [-c archiv-dir] [-s] [-n] [-f] [files]\n",
			       progname);
		exit(-1);
	}
	if (mapdir != NULL)
		archdir(mapdir);

	if (savehdrs > 0)
		savefp = fopen(savfil, "a");

	if (mapgrp != NULL) {
		domaps(mapgrp);
	} else if (optind < argc) {
		/* some arguments, use them as filenames */
		mapgrp = NULL;
		seqfil = NULL;
		for(x = optind; x < argc; x++)
			(void) mkmaps((FILE *) NULL, argv[x]);
	} else {
		/* act as filter for stdin */
		(void) mkmaps(stdin, "-");
	}

#ifdef DEBUG
	mal_verify(1);
#endif
	return 0;
}

/* compar routine called by the sort on the (numeric) filenames */
int
numsort(d1, d2)
char *d1;
char *d2;
{
	return atoi((*(struct direct **)d1)->d_name) - 
		atoi((*(struct direct **)d2)->d_name);
}

static long artmin;
static int artlen;

/* select routine for scandir(), returns 1 if the file is worth considering */
int
fselect(dp)
struct direct *dp;
{
	register char *cp = dp->d_name;

	if (*cp == '.' && (cp[1] == '\0' ||
			   (cp[1] == '.' && cp[2] == '\0')))
		return 0;
	while(*cp && isascii(*cp) && isdigit(*cp))
		cp++;
	if (*cp != '\0')
		return 0;
	if (cp - dp->d_name >= artlen) {
		logerr("Name too long: %s", dp->d_name);
		return 0;
	}
	if (artmin >= atoi(dp->d_name))
		return 0;
	return 1;
}

void
domaps(grp)
char	*grp;
{
	FILE	*seqsd;
#ifndef DEBUG
	char	nbuf[MAXSTR];
#else
	char	*nbuf = malloc(MAXSTR);
#endif
	char	*nptr;
	struct direct **filst;
	int	nfils, x;

	artmin = 0;
	if ((seqsd = fopen(seqfil, "r+")) != NULL) {
		if (fgets(nbuf, MAXSTR, seqsd) != NULL)
			artmin = atoi(nbuf);
	} else if ((seqsd = fopen(seqfil, "w")) == NULL)
		error("Could not open seq file for ", "");

	(void) strcpy(nbuf, grp);
	nptr = nbuf + strlen(nbuf);
	*nptr++ = '/';
	*nptr = '\0';
	artlen = MAXSTR - (nptr - nbuf) - 1;

	if ((nfils = scandir(grp, &filst, fselect, numsort)) == -1)
		error("scandir failed", "");

	for (x = 0; x < nfils; x++) {
		(void) strcpy(nptr, filst[x]->d_name);
		if (mkmaps((FILE *) NULL, nbuf) < 0)
			continue;
		if (seqsd != NULL) {
			(void) fseek(seqsd, 0L, 0);
			(void) fputs(filst[x]->d_name, seqsd);
			(void) putc('\n', seqsd);
			if (fflush(seqsd) != 0)
				warning("error flushing seq file %s", seqfil);
		}
	}
	if (seqsd != NULL && fclose(seqsd) != 0)
		warning("error closing seq file %s", seqfil);
}

int
mkmaps(isd, file)
FILE	*isd;
char	*file;
{
#define	SEARCH		1
#define	INAMAP		2
#define	SKIPPING	3

	struct stat stbuf;
#ifndef DEBUG
	char	buf[MAXSTR], tofil[MAXSTR], ntofil[MAXSTR + 32], delim[MAXSTR];
#else
	char	*buf = malloc(MAXSTR);
	char	*tofil = malloc(MAXSTR);
	char	*ntofil = malloc(MAXSTR + 32);
	char	*delim = malloc(MAXSTR);
#endif
	int	state = SEARCH, sizdel = 0, n, nlines = 0, sedch = 0;
	FILE	*osd = NULL;
	register char *morep = buf + MAXSTR - 2;
#define TEMPLATE "uumapXXXXXX"
	char tfname[sizeof TEMPLATE];

	if (isd == NULL && (isd = fopen(file, "r")) == NULL) {
		warning("Could not open %s. Skipping...", file);
		return -1;
	}
	if (fstat(fileno(isd), &stbuf) != 0) {
		warning("Could not stat %s", file);
		return -1;
	}
	if ((stbuf.st_mode & S_IFMT) == S_IFDIR)
		return -1;
	strcpy(tfname, TEMPLATE);
	if (mktemp(tfname) == NULL) {
		logerr("mktemp(%s) failed", tfname);
		return -1;
	}

	log("Unarchive %s", file);
	for(;;) {
		*morep = '\0';
		nlines++;
		if (fgets(buf, MAXSTR, isd) == NULL)
			break;
		/*
		 * This is non-portable since an fgets() implementation
		 * may mess with the byte even if the buffer doesn't
		 * fill up.  However, it's unlikely.
		 */
		if (*morep != '\0' && *morep != '\n') {
			if (state != INAMAP) {	/* silently skip line */
				int c;

				if (savefp)
					fprintf(savefp, "%s", buf);
				while((c = getc(isd)) != EOF && c != '\n') {
					if (savefp)
						putc(c, savefp);
				}
				continue;
			}
			/* trouble, bail out */
			sprintf(buf, "%d greater than %d chars", nlines, 
				MAXSTR);
			logerr("line %s", buf);
			(void) fclose(isd);
			if (osd != NULL)
				(void) fclose(osd);
			return -1;
		}
		switch (state) {
		case SEARCH:
			sedch = gotcmd(buf, tofil, MAXSTR, delim, MAXSTR);
			if (sedch >= 0) {
				sizdel = strlen(delim);
				if (sizdel <= 0) {
					if (savefp)
						fputs(buf, savefp);
					break;
				}
				if (badname(tofil)) {
					logerr("bad name, skipping `%s'\n",tofil);
					state = SKIPPING;
					continue;
				}
				state = INAMAP;
				log("%s", tofil);
				if ((osd = fopen(tfname, "w")) == NULL) {
					logerr("Could not open %s", tfname);
					state = SKIPPING;
				}
			} else if (savefp) {	
				fputs(buf, savefp);
			}
			break;
		case SKIPPING:
			if (STREQN(buf, delim, sizdel))
				state = SEARCH;
			break;
		case INAMAP:
			if (STREQN(buf, delim, sizdel)) {
				state = SEARCH;
				if (osd != NULL && fclose(osd) != 0)
					error("fclose(%s) failed", tfname);
				osd = NULL;
				n = 1;
				if (stat(tofil, &stbuf) != -1) {
					do {
						sprintf(ntofil,
							"%s-%d", tofil, n++);
					} while (link(tofil, ntofil) < 0
						&& errno == EEXIST);
				}
				/* rename()? */
				if (unlink(tofil) < 0 && errno != ENOENT)
					error("couldn't unlink %s", tofil);
				if (link(tfname, tofil) < 0)
					error("couldn't link tmpfile to %s",
					      tofil);
				(void) unlink(tfname);
			} else
				fputs(buf + (*buf == sedch ? 1: 0), osd);
			break;
		}
	}

	if (state != SEARCH) {
		char ebuf[MAXSTR];
		
		sprintf(ebuf, "Read/sync error on %s (state %d, offset %ld, '%*s')",
			file, state, ftell(isd), sizdel, delim);
		logerr("%s", ebuf);
	}

	if (osd != NULL && fclose(osd) != 0)
		error("fclose(%s) failed", tofil);
	(void) fclose(isd);
	(void) unlink(tfname);

	return 0;

#undef	SEARCH
#undef	INAMAP
#undef	SKIPPING
}

/*
 * gotcmd
 *
 * recognize
 *	cat << DELIM > OUT
 * or
 *	sed 's/^?//' > OUT << DELIM
 * and several legal shell variants thereof.  OUT and DELIM cannot
 * be plain cat or sed.  But they can be used as such if quoted.
 *
 */
int
gotcmd(buf, tofil, tofilln, delim, delimln)
register char	*buf;
char	*tofil, *delim;
int	tofilln, delimln;
{
	register int	cmdtype = 0;
	int gotflags = 0;
#define GOT_CMD		(1 << 1)
#define GOT_ARGS	(1 << 2)
#define GOT_DELIM	(1 << 3)
#define GOT_FILE	(1 << 4)
#define GOT_ALL		(GOT_CMD | GOT_ARGS | GOT_DELIM | GOT_FILE)
	int sedchar = -1;
#define CMD_CAT	1
#define CMD_SED	2
#define MKDIR	"mkdir"
#define MKDIRLEN	(sizeof(MKDIR)-1)

#define ISDELIM(c) ((c) == '>' || (c) == '<' || (isascii(c) && isspace(c)))

	tofil[0] = delim[0] = '\0';
	for(;;) {
		/* gobble whitespace */
		while(*buf != '\0' && isascii(*buf) && isspace(*buf))
			buf++;
		if (*buf == '\0')
			break;
		if (*buf == '>') {
			buf++;
			buf = getwrd(buf, tofil, tofilln);
			if (tofil[0] != '\0')
				gotflags |= GOT_FILE;
		} else if (*buf == '<' && buf[1] == '<') {
			buf += 2;
			buf = getwrd(buf, delim, delimln);
			if (delim[0] != '\0')
				gotflags |= GOT_DELIM	;
		} else if (cmdtype == 0 && *buf == 'c' && buf[1] == 'a' &&
			   buf[2] == 't' && ISDELIM(buf[3])) {
			buf += 3;
			cmdtype = CMD_CAT;
			gotflags |= (GOT_CMD | GOT_ARGS);
			sedchar = '\0';
		} else if (cmdtype == 0 && *buf == 's' && buf[1] == 'e' &&
			   buf[2] == 'd' && ISDELIM(buf[3])) {
			buf += 3;
			cmdtype = CMD_SED;
			gotflags |= GOT_CMD;
		} else if (cmdtype == CMD_SED && (gotflags & GOT_ARGS) == 0) {
			/* must be the sed command */
			buf = getsed(buf, &sedchar);
			if (sedchar >= 0)
				gotflags |= GOT_ARGS;
		} else if (cmdtype == 0 && STREQN(buf, MKDIR, MKDIRLEN) &&
			   ISDELIM(buf[MKDIRLEN])) {
			char dirbuf[MAXSTR];

			dirbuf[0] = '\0';
			buf += MKDIRLEN;
			buf = getwrd(buf, dirbuf, sizeof(dirbuf));
			if (dirbuf[0] == '\0') {
				logerr("skipping mkdir with no argument", "");
			} else if (nosubdirs > 0) {
				logerr("ignoring mkdir %s", dirbuf);
			} else if (badname(dirbuf) > 0) {
				logerr("skipping unsafe directory %s", dirbuf);
			} else if (mkdir(dirbuf, 0777) < 0) {
				warning("couldn't mkdir(%s)", dirbuf);
			} else {
				log("making directory %s", dirbuf);
			}
			return -1;
		} else if (gotflags == GOT_ALL && buf[0] == '&' &&
			   buf[1] == '&' &&
			   (buf[2] == '\0' || buf[2] == '\n')) {
			/*
			 * some shar deviant puts && at the end of the
			 * commands.  Yeesh!  If you need that much
			 * reliability, use ar or tar.
			 */
			buf += 2;
		} else
			return -1;
	}
	if (gotflags == GOT_ALL)
		return(sedchar);
	return -1;
}

/*
 * getwrd copies a word from src to buf, following shell quoting fairly
 * closely.  The main difference is that newline is treated as end
 * of word, even if quoted.
 */
char *
getwrd(src, buf, maxlen)
char	*src, *buf;
int	maxlen;
{
	register char	*cp = src;
	register char	*ptr = buf;
	register char	*end = buf + maxlen - 1;
	register int	inquotes = 0;
	int slashquote = 0;

	while(*cp != '\0' && isascii(*cp) && isspace(*cp))
		cp++;
	while (*cp != '\0') {
		if (*cp == '\'' || *cp == '"') {
			if (inquotes == 0) {
				inquotes = *cp++;
				continue;
			} else if (*cp == inquotes) {
				inquotes = 0;
				cp++;
				continue;
			}
		}
		if (*cp == '\\') {
			slashquote = 1;
			if (inquotes == 0)
				cp++;
			else if (inquotes == '"' && (*(cp + 1) == '"' ||
						     *(cp + 1) == '\\'))
				cp++;
			else
				slashquote = 0;
		}
		/*
		 * if someone tries sticking in a newline, or the word
		 * is too long, we quit.  !! should probably return an
		 * error instead of truncating.
		 */
		if (*cp == '\0' || *cp == '\n' ||
		    (slashquote == 0 && inquotes == 0 && ISDELIM(*cp)) ||
		    ptr == end)
			break;
		*ptr++ = *cp++;
		slashquote = 0;
	}
	*ptr = '\0';
	if (debug)
		fprintf(stderr, "getwrd extracted `%s', returning `%s'\n",
			buf, cp);
	return cp;
}

/*
 * extracts the X from a sed command of the form s/^X// which is what
 * all shars use
 */
char *
getsed(cp, sedchp)
char *cp;
int *sedchp;
{
	char sedbuf[MAXSTR];

	*sedchp = -1;
	cp = getwrd(cp, sedbuf, sizeof sedbuf);
	if (sedbuf[0] == '\0') {	/* no arg means sed == cat */
		*sedchp = 0;
	} else if (sedbuf[0] == 's' && sedbuf[1] == '/' && sedbuf[2] == '^' &&
	    sedbuf[3] != '\0' && sedbuf[3] != '/' && sedbuf[3] != '\\' &&
	    sedbuf[4] == '/' && sedbuf[5] == '/' && sedbuf[6] == '\0') {
		*sedchp = sedbuf[3];
	}
	return cp;
}

/* checks that filename is relative, doesn't contain .. */
int
badname(s)
char *s;
{
	register char *cp = s;
	register char c, lastc = '\0';
	char ebuf[MAXSTR];

	if (*cp == '/') {
		logerr("absolute pathname %s", cp);
		return 1;
	}
	while((c = *cp++) != '\0') {
		if (!isascii(c) || !isprint(c)) {
			sprintf(ebuf, "0x%x", c);
			logerr("unprintable character %s", ebuf);
			return 1;
		}
		if (c == '.' && lastc == '.') {
			logerr(".. in name %s", s);
			return 1;
		}
		lastc = c;
		if (strchr(badchars, c) != NULL) {
			sprintf(ebuf, "%c", c);
			logerr("bad character %s", ebuf);
			return 1;
		}
	}
	return 0;
}

void
log(fmt, s)
char *fmt;
char *s;
{
	(void) printf(fmt, s);
	(void) putc('\n', stdout);
}

void
logerr(fmt, s)
char *fmt;
char *s;
{
	(void) fprintf(stderr, fmt, s);
	(void) putc('\n', stderr);
	(void) fflush(stderr);
}

void
archdir(dirname)
char *dirname;
{
	struct stat stbuf;

	if (stat(dirname, &stbuf) < 0) {
		if (errno != ENOENT)
			error("couldn't stat directory %s", mapdir);
		/* !! perhaps we should do a mkdir -p? */
		if (mkdir(dirname, 0777) < 0)
			error("couldn't mkdir directory %s", mapdir);
	}
	if (chdir(mapdir) < 0)
		error("Could not change directory to %s", mapdir);
}
